from attrdict import AttrDict
from collections import OrderedDict, defaultdict
from datetime import timedelta

from django.db import transaction
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _l
from django.utils.timezone import now

from pymess.config import settings
from pymess.config import get_router, get_backend, get_default_sender_backend_name
from pymess.utils import fullname


class BaseController:
    """
    Base class of Controller. Any type of communication requires Controller derived from this class.
    Many backends of the same type of communication can be supported by one controller which chooses the right backend
    implementation based on the recipient
    """

    model = None
    backend_type_name = None

    def __init__(self):
        self._loaded_backends = {}

    def get_backend(self, recipient):
        backend_name = self.router.get_backend_name(recipient) or get_default_sender_backend_name(self.backend_type_name)
        if backend_name not in self._loaded_backends:
            self._loaded_backends[backend_name] = get_backend(self.backend_type_name, backend_name)
        return self._loaded_backends[backend_name]

    @cached_property
    def router(self):
        return get_router(self.backend_type_name)

    def get_waiting_or_retry_messages(self):
        """
        Return queryset of waiting messages to send
        """
        return self.model.objects.filter(state__in={self.model.STATE.WAITING, self.model.STATE.ERROR_RETRY})

    def is_turned_on_batch_sending(self):
        return False

    def publish_or_retry_message(self, message):
        backend = self.get_backend(recipient=message.recipient)
        if (message.number_of_send_attempts > backend.get_batch_max_number_of_send_attempts()
            or message.created_at < now() - timedelta(seconds=self.get_batch_max_seconds_to_send())):
            backend._set_message_as_failed(message)
            return False
        else:
            backend.publish_message(message)
            return True

    @transaction.atomic
    def send(self, recipient, content, related_objects=None, tag=None, template=None, send_immediately=False,
             message_backend=None, **kwargs):
        """
        Send message with the text content to the phone number (recipient)
        :param recipient: email or phone number of the recipient
        :param content: text content of the message
        :param related_objects: list of related objects that will be linked with the message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was create
        :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
        :param message_backend: message backend instance
        :param kwargs: extra attributes that will be stored to the message
        """
        backend = message_backend or self.get_backend(recipient)
        message = self.create_message(recipient=recipient, content=content, related_objects=related_objects, tag=tag,
                                      template=template, **kwargs)
        if send_immediately or not self.is_turned_on_batch_sending():
            backend.publish_message(message)
        return message

    def get_batch_max_seconds_to_send(self):
        """
        Return max timeout in seconds to send message
        """
        raise NotImplementedError

    def get_batch_size(self):
        """
        Return number of messages sent in batch
        """
        raise NotImplementedError

    def create_message(self, recipient, content, related_objects, tag, template,
                       priority=settings.DEFAULT_MESSAGE_PRIORITY, **kwargs):
        """
        Create message which will be logged in the database.
        :param recipient: email or phone number of the recipient
        :param content: content of the message
        :param related_objects: list of related objects that will be linked with the message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param priority: priority of sending message 1 (highest) to 3 (lowest)
        :param kwargs: extra attributes that will be saved with the message
        """
        message = self.model.objects.create(
            recipient=recipient,
            content=content,
            tag=tag,
            template=template,
            template_slug=template.slug if template else None,
            priority=priority,
            **kwargs
        )
        if related_objects:
            message.related_objects.create_from_related_objects(*related_objects)
        return message

    def _get_backend_messages_map(self, messages):
        backends_messages_map = defaultdict(list)
        for message in messages:
            backend = self.get_backend(recipient=message.recipient)
            backends_messages_map[backend].append(message)
        return backends_messages_map

    def bulk_send_messages(self, messages):
        """
        Sends more messages together. If concrete backend provides send more messages at once the method
        can be overridden
        :param messages: list of messages
        """
        for backend, messages_for_backend in self._get_backend_messages_map(messages).items():
            backend.publish_messages(messages_for_backend)

    def bulk_send(self, recipients, content, related_objects=None, tag=None, template=None, **kwargs):
        """
        Send more messages in one bulk
        :param recipients: list of emails or phone numbers of recipients
        :param content: text content of the messages
        :param related_objects: list of related objects that will be linked with the message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was create
        :param kwargs: extra attributes that will be stored with messages
        """
        with transaction.atomic():
            messages = [
                self.create_message(recipient, content, related_objects, tag, template, **kwargs)
                for recipient in recipients
            ]
        self.bulk_send_messages(messages)
        return messages


class BaseBackend:

    config = AttrDict()

    def __init__(self, config=None):
        self.config = AttrDict({**self.config, **(config or {})})

    def _get_extra_sender_data(self):
        """
        Gets arguments that will be saved with the message in the extra_sender_data field
        """
        return {}

    def get_extra_message_kwargs(self):
        """
        Gets model message kwargs that will be saved with the message
        """
        return {}

    def _update_message(self, message, extra_sender_data=None, **kwargs):
        """
        Method for updating state of the message
        :param message: message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param kwargs: changed object kwargs
        """
        extra_sender_data = {
            **self._get_extra_sender_data(),
            **({} if extra_sender_data is None else extra_sender_data)
        }
        message.change_and_save(
            backend=fullname(self),
            extra_sender_data=extra_sender_data,
            **kwargs
        )

    def _update_message_after_sending(self, message, extra_sender_data=None, **kwargs):
        """
        Method for updating state of the message after it was sent
        :param message: message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param kwargs: changed object kwargs
        """
        self._update_message(
            message,
            extra_sender_data,
            number_of_send_attempts=message.number_of_send_attempts + 1,
            **kwargs
        )

    def _update_message_after_sending_error(self, message, extra_sender_data=None, state=None, **kwargs):
        """
        Method for updating state of the message after it was send with error result
        :param message: message object
        :param extra_sender_data: extra data that will be saved to the extra_sender_data field
        :param state: error state of the message
        :param kwargs: changed object kwargs
        """

        number_of_send_attempts = message.number_of_send_attempts + 1

        if not state:
            state = (
                message.STATE.ERROR if (
                    number_of_send_attempts > self.get_batch_max_number_of_send_attempts()
                    or not self.get_retry_sending()
                ) else message.STATE.ERROR_RETRY
            )

        self._update_message(
            message,
            extra_sender_data,
            state=state,
            number_of_send_attempts=number_of_send_attempts,
            **kwargs
        )

    def _set_message_as_failed(self, message):
        """
        Method for updating state of the message to the final error state
        :param message: message object
        """
        self._update_message(
            message,
            state=message.STATE.ERROR,
        )

    def publish_message(self, message):
        """
        Send the message
        :param message: SMS message
        """
        raise NotImplementedError

    def publish_messages(self, messages):
        """
        Send bulk of messages at once
        :param messages: list of SMS message
        """
        return [self.publish_message(message) for message in sorted(messages, key=lambda m: m.priority)]

    def get_batch_max_number_of_send_attempts(self):
        """
        Return number attempts to send message
        """
        raise NotImplementedError

    def get_retry_sending(self):
        """
        Return True if message should be retried if sending fails
        """
        raise NotImplementedError


def send_template(recipient, slug, context_data, related_objects=None, tag=None, template_model=None, **kwargs):
    """
    Helper for building and sending message from a template.
    :param recipient: email or phone number of the recipient
    :param slug: slug of a template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
    :param tag: string mark that will be saved with the message
    :param template_model: template model instance
    :param kwargs: extra attributes that will be stored with message
    :return: dialer message object or None if template cannot be sent
    """

    assert template_model is not None, _l('template_model cannot be None')

    return template_model.objects.get(slug=slug).send(
        recipient,
        context_data,
        related_objects=related_objects,
        tag=tag,
        **kwargs
    )


def send(recipient, content, related_objects=None, tag=None, message_controller=None, **kwargs):
    """
    Helper for sending message.
    :param recipient: email or phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :param message_controller: controller sender instance
    :return: True if message was successfully sent or False if message is in error state
    """
    return message_controller.send(
        recipient,
        content,
        related_objects=related_objects,
        tag=tag,
        **kwargs
    ).failed



from chamber.exceptions import PersistenceException

from pymess.backend import BaseBackend, send_template as _send_template, send as _send, BaseController
from pymess.config import (
    CONTROLLER_TYPES, get_push_notification_template_model, is_turned_on_push_notification_batch_sending, settings
)
from pymess.models import PushNotificationMessage


class PushNotificationController(BaseController):
    """Controller class for push notifications delegating message to correct push notification backend"""

    model = PushNotificationMessage
    backend_type_name = CONTROLLER_TYPES.PUSH_NOTIFICATION

    class PushNotificationSendingError(Exception):
        pass

    def get_batch_max_seconds_to_send(self):
        return settings.PUSH_NOTIFICATION_BATCH_MAX_SECONDS_TO_SEND

    def get_batch_size(self):
        return settings.PUSH_NOTIFICATION_BATCH_SIZE

    def is_turned_on_batch_sending(self):
        return is_turned_on_push_notification_batch_sending()

    def create_message(self, recipient, content, related_objects, tag, template,
                       priority=settings.DEFAULT_MESSAGE_PRIORITY, **kwargs):
        try:
            notification = super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                template=template,
                tag=tag,
                priority=priority,
                **kwargs
            )
            return notification
        except PersistenceException as ex:
            raise self.PushNotificationSendingError(str(ex))


class PushNotificationBackend(BaseBackend):

    def get_batch_max_number_of_send_attempts(self):
        return settings.PUSH_NOTIFICATION_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS

    def get_retry_sending(self):
        return settings.PUSH_NOTIFICATION_RETRY_SENDING and is_turned_on_push_notification_batch_sending()


def send_template(recipient, slug, context_data, related_objects=None, tag=None, send_immediately=None):
    """
    Helper for building and sending push notification message from a template.
    :param recipient: push notification recipient
    :param slug: slug of a push notifiaction template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the push notification using generic
        relation
    :param tag: string mark that will be saved with the message
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :return: Push notification message object or None if template cannot be sent
    """
    return _send_template(
        recipient=recipient,
        slug=slug,
        context_data=context_data,
        related_objects=related_objects,
        tag=tag,
        template_model=get_push_notification_template_model(),
        send_immediately=send_immediately
    )


def send(recipient, content, related_objects=None, tag=None, send_immediately=False, **kwargs):
    """
    Helper for sending push notification.
    :param recipient: push notification recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :return: True if push notification was successfully sent or False if message is in error state
    """
    return _send(
        recipient=recipient,
        content=content,
        related_objects=related_objects,
        tag=tag,
        message_controller=PushNotificationController(),
        send_immediately=send_immediately,
        **kwargs
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2020-01-08 23:02
from __future__ import unicode_literals

from django.db import migrations, models
from django.db.models import Func, F, Value


def copy_external_id_from_extra_data(apps, schema_editor):
    EmailMessage = apps.get_model('pymess', 'EmailMessage')
    EmailMessage.objects.all().update(
        external_id=Func(F('extra_sender_data'), Value('^.*"_id":"([^"]*)".*$'), function='substring')
    )


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0013_migration'),
    ]

    operations = [
        migrations.RunPython(copy_external_id_from_extra_data),
    ]



import codecs
import os

from chamber.shortcuts import change_and_save

from pymess.config import settings


class SyncEmailTemplate:

    def __init__(self, template_slug):
        self.template_slug = template_slug

    def __call__(self, apps, schema_editor):
        directory = os.path.join(settings.EMAIL_HTML_DATA_DIRECTORY)

        with codecs.open(os.path.join(directory, '{}.html'.format(self.template_slug)), 'r',
                         encoding='utf-8-sig') as file:
            template_obj = apps.get_model(*settings.EMAIL_TEMPLATE_MODEL.split('.')).objects.get(
                slug=self.template_slug
            )
            change_and_save(template_obj, body=file.read())




import re
from attrdict import AttrDict

from django.utils import timezone as tz
from django.utils.translation import ugettext as _

from pymess.backend.dialer import DialerBackend
from pymess.config import settings
from pymess.models import DialerMessage
from pymess.utils.logged_requests import generate_session


class DaktelaDialerBackend(DialerBackend):
    """
    Dialer backend implementing Daktela service https://www.daktela.com/api/v6/models/campaignsrecords
    """

    SESSION_SLUG = 'pymess-daktela_autodialer'
    config = AttrDict({
        'ACCESS_TOKEN':  None,
        'AUTODIALER_QUEUE': None,
        'PREDICTIVE_QUEUE': None,
        'URL': None,
        'STATES_MAPPING': {
            '0': 0,
            '1': 1,
            '2': 2,
            '3': 3,
            '4': 4,
            '5': 5,
            '6': 6,
        },
        'TIMEOUT': 5,  # 5s
    })

    def _get_dialer_api_url(self, name=None):
        name = '/{name}'.format(name=name) if name else ''

        return '{base_url}{name}.json?accessToken={access_token}'.format(
            base_url=self.config.URL, name=name, access_token=self.config.ACCESS_TOKEN,
        )

    def _get_autodialer_message_states(self, resp_json, state_mapped, resp_message_state):
        custom_fields = resp_json['result']['customFields']
        tts_processed = custom_fields['ttsprocessed'][0]
        whole_message_heard = ('whole_message_heard' in custom_fields and custom_fields['whole_message_heard']
                               and custom_fields['whole_message_heard'][0] == 'Yes')
        if state_mapped == DialerMessage.STATE.ANSWERED_PARTIAL and whole_message_heard:
            resp_message_state = str(DialerMessage.STATE.ANSWERED_COMPLETE)
        if state_mapped == DialerMessage.STATE.DONE and tts_processed == '0':
            resp_message_state = str(DialerMessage.STATE.NOT_ASSIGNED)
        tts_processed = resp_json['result']['customFields']['ttsprocessed'][0]
        is_final_state = resp_json['result']['action'] == '5' and tts_processed == '1'
        message_state = self.config.STATES_MAPPING[resp_message_state]
        return is_final_state, message_state

    def _update_dialer_states(self, messages):
        """
        Method uses Daktela API to get info about autodialer call status
        :param messages: list of dialer messages to update
        """
        for message in messages:
            name = message.extra_data['name']
            client_url = self._get_dialer_api_url(name)
            response = generate_session(
                slug=self.SESSION_SLUG,
                related_objects=(message,),
                timeout=self.config.TIMEOUT
            ).get(client_url)
            resp_json = response.json()
            if response.status_code != 200 and resp_json.get('error'):
                self._update_message_state_with_error(message, error_message=resp_json.get('error'))
                continue

            message.extra_data.update({
                'name': name,
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })
            resp_message_state = resp_json['result']['statuses'][0]['name'] if len(
                resp_json['result']['statuses']) else resp_json['result']['action']
            state_mapped = self.config.STATES_MAPPING[resp_message_state]
            message_error = resp_json['error'] if len(resp_json['error']) else None

            if message.is_autodialer:
                is_final_state, message_state = self._get_autodialer_message_states(resp_json, state_mapped,
                                                                                    resp_message_state)
            else:
                message_state = state_mapped
                is_final_state = resp_json['result']['action'] == '5'

            try:
                self._update_message(
                    message,
                    state=message_state,
                    error=message_error,
                    extra_data=message.extra_data,
                    is_final_state=is_final_state,
                )
            except Exception as ex:
                self._update_message_state_with_error(message, error_message=ex)
                # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
                # and log them into database. Re-raise exception causes transaction rollback (lost of information about
                # exception).

    def _update_message_state_with_error(self, message, error_message):
        is_final_state = message.number_of_status_check_attempts >= settings.DIALER_NUMBER_OF_STATUS_CHECK_ATTEMPTS
        message_kwargs = {
            'state': DialerMessage.STATE.ERROR_UPDATE,
            'error': ', '.join(error_message) if isinstance(error_message, list) else str(error_message),
            'is_final_state': is_final_state,
        }
        if not is_final_state:
            message_kwargs['number_of_status_check_attempts'] = message.number_of_status_check_attempts + 1
        self._update_message(
            message,
            **message_kwargs
        )

    def publish_message(self, message):
        """
        Method uses Daktela API for sending dialer message
        :param message: dialer message
        """
        client_url = self._get_dialer_api_url()
        try:
            payload = {
                'queue': (self.config.AUTODIALER_QUEUE if message.is_autodialer
                          else self.config.PREDICTIVE_QUEUE),
                'number': re.sub(r'^\+', '00', message.recipient),
                'customFields': {
                    'mall_pay_text': [
                        message.content,
                    ],
                    'ttsprocessed': [
                        0,
                    ],
                },
            }
            if message.is_autodialer:
                payload['action'] = 5
            custom_fields = message.extra_data.get('custom_fields')
            if custom_fields:
                payload['customFields'].update(**custom_fields)

            response = generate_session(
                slug=self.SESSION_SLUG,
                related_objects=(message,),
                timeout=self.config.TIMEOUT
            ).post(
                client_url,
                json=payload,
            )
            resp_json = response.json()

            message.extra_data.update({
                'name': resp_json['result']['name'],
                'daktela_action': resp_json['result']['action'],
                'daktela_statuses': resp_json['result']['statuses'],
            })

            error_message = resp_json.get('error')
            if error_message:
                self._update_message_after_sending_error(
                    message,
                    error=', '.join(error_message) if isinstance(error_message, list) else str(error_message),
                    state=DialerMessage.STATE.ERROR
                )
            else:
                self._update_message_after_sending(
                    message,
                    state=DialerMessage.STATE.READY,
                    sent_at=tz.now(),
                    extra_data=message.extra_data,
                )
        except Exception as ex:
            self._update_message_after_sending_error(
                message,
                error=str(ex)
            )
            # Do not re-raise caught exception. We do not know exact exception to catch so we catch them all
            # and log them into database. Re-raise exception causes transaction rollback (lost of information about
            # exception).



import datetime
import logging

from django.core.management.base import BaseCommand
from django.db.models import F, Q
from django.utils.timezone import now

from chamber.utils.transaction import atomic_with_signals

from pymess.backend.emails import EmailController
from pymess.config import settings


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Command for pulling info of sent e-mails.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.touched_message_pks = set()
        self.updated_messages = set()

    def _get_messages_queryset_to_update(self, email_controller):
        delay = datetime.timedelta(seconds=settings.EMAIL_PULL_INFO_DELAY_SECONDS)
        return email_controller.model.objects.filter(
            # start_time = last_webhook_received_at + delay
            # info_changed_at < start_time
            # info_changed_at < last_webhook_received_at + delay
            Q(info_changed_at__isnull=True) | Q(info_changed_at__lt=F('last_webhook_received_at') + delay),
            last_webhook_received_at__lt=now() - delay,
            sent_at__gt=now() - datetime.timedelta(seconds=settings.EMAIL_PULL_INFO_MAX_TIMEOUT_FROM_SENT_SECONDS)
        ).order_by('-sent_at')

    @atomic_with_signals
    def _pull_message_info(self, email_controller):
        message = self._get_messages_queryset_to_update(email_controller).exclude(
            pk__in=self.touched_message_pks
        ).select_for_update().first()

        if not message:
            return False

        self.touched_message_pks.add(message.pk)
        email_controller.get_backend(message.recipient).pull_message_info(message)
        self.updated_messages.add(message.pk)
        return True

    def _print_result(self, title, message_pks):
        if message_pks:
            self.stdout.write('{}: {} ({})'.format(title, len(message_pks), ', '.join((str(pk) for pk in message_pks))))
        else:
            self.stdout.write('{}: {}'.format(title, len(message_pks)))

    def handle(self, *args, **options):
        email_controller = EmailController()

        for _ in range(settings.EMAIL_PULL_INFO_BATCH_SIZE):
            try:
                if not self._pull_message_info(email_controller):
                    break
            except Exception as ex:
                # One message error should not stop the whole cycle
                logger.exception(ex)

        self._print_result('updated messages', self.updated_messages)
        self._print_result('failed messages', self.touched_message_pks - self.updated_messages)



from functools import reduce
from operator import or_ as OR

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Q
from django.db.models.functions import Cast
from django.template import Context, Template
from django.template.exceptions import TemplateDoesNotExist, TemplateSyntaxError
from django.utils.translation import ugettext_lazy as _

from chamber.models import SmartModel
from chamber.utils.datastructures import ChoicesNumEnum

from pymess.config import settings


class RelatedObjectQueryset(models.QuerySet):

    def annotate_object_pks(self, model_class):
        pk_field = model_class._meta.pk
        if isinstance(pk_field, models.AutoField):
            pk_field = models.IntegerField()
        return self.filter(
            content_type=ContentType.objects.get_for_model(model_class)
        ).annotate(
            object_pk=Cast('object_id', output_field=pk_field)
        )


class RelatedObjectManager(models.Manager):

    def create_from_related_object(self, related_object):
        return self.create(
            object_id=related_object.pk,
            content_type=ContentType.objects.get_for_model(related_object)
        )

    def create_from_related_objects(self, *related_objects):
        return [
            self.create_from_related_object(related_object)
            for related_object in related_objects
        ]

    def _get_related_object_pks(self, *related_objects):
        return [related_object.pk for related_object in related_objects]

    def _get_related_objects_qs_kwargs(self, *related_objects):
        return reduce(OR, (
            Q(**{
                'object_id': obj.pk,
                'content_type': ContentType.objects.get_for_model(obj)
            }) for obj in related_objects
        ))

    def update_related_objects(self, *related_objects, limit_to_model=None):
        delete_qs = self.all()
        if limit_to_model is not None:
            delete_qs = delete_qs.filter(content_type=ContentType.objects.get_for_model(limit_to_model))
        if related_objects:
            delete_qs = delete_qs.exclude(self._get_related_objects_qs_kwargs(*related_objects))
        delete_qs.delete()
        return self.get_or_create_from_related_objects(*related_objects)

    def get_or_create_from_related_object(self, related_object):
        return self.get_or_create(
            object_id=related_object.pk,
            content_type=ContentType.objects.get_for_model(related_object)
        )

    def get_or_create_from_related_objects(self, *related_objects):
        return [
            self.get_or_create_from_related_object(related_object)
            for related_object in related_objects
        ]

    def get_related_objects_by_model(self, model_class):
        return model_class.objects.filter(pk__in=self.annotate_object_pks(model_class).values('object_pk'))

    def filter_from_related_objects(self, *related_objects):
        if hasattr(self, 'instance'):
            template_filter = (Q(template=self.instance) | Q(template__isnull=True))
        else:
            template_filter = Q(template__isnull=True)
        return self.model.objects.filter(template_filter).filter(self._get_related_objects_qs_kwargs(*related_objects))


class MessageQueryset(models.QuerySet):

    def filter_related_object(self, related_object):
        return self.filter(
            related_objects__object_id=str(related_object.pk),
            related_objects__content_type=ContentType.objects.get_for_model(related_object)
        )

    def filter_related_objects(self, related_objects_queryset):
        return self.filter(
            related_objects__object_id__in=related_objects_queryset.annotate(
                object_str_pk=Cast('pk', output_field=models.TextField())
            ).values('object_str_pk'),
            related_objects__content_type=ContentType.objects.get_for_model(related_objects_queryset.model)
        )


class MessageManager(models.Manager):

    def was_sent(self, template_slug, related_object):
        return self.filter(
            template_slug=template_slug,
        ).filter_related_object(related_object).exists()


class BaseMessage(SmartModel):

    STATE = ChoicesNumEnum()

    sent_at = models.DateTimeField(verbose_name=_('sent at'), null=True, blank=True, editable=False)
    recipient = models.CharField(verbose_name=_('recipient'), null=False, blank=False, max_length=20, db_index=True)
    content = models.TextField(verbose_name=_('content'), null=False, blank=False)
    template_slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=True, blank=True, editable=False,
                                     db_index=True)
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False)
    backend = models.CharField(verbose_name=_('backend'), null=True, blank=True, editable=False, max_length=250)
    error = models.TextField(verbose_name=_('error'), null=True, blank=True, editable=False)
    extra_data = models.JSONField(verbose_name=_('extra data'), null=True, blank=True, editable=False,
                                  encoder=DjangoJSONEncoder)
    extra_sender_data = models.JSONField(verbose_name=_('extra sender data'), null=True, blank=True, editable=False,
                                         encoder=DjangoJSONEncoder)
    tag = models.SlugField(verbose_name=_('tag'), null=True, blank=True, editable=False)
    number_of_send_attempts = models.PositiveIntegerField(verbose_name=_('number of send attempts'), null=False,
                                                          blank=False, default=0)
    priority = models.PositiveSmallIntegerField(verbose_name=_('priority'), null=False, blank=False,
                                                default=settings.DEFAULT_MESSAGE_PRIORITY)

    objects = MessageManager.from_queryset(MessageQueryset)()

    def __str__(self):
        return self.recipient

    class Meta:
        abstract = True
        ordering = ('-created_at',)


class BaseRelatedObject(SmartModel):

    content_type = models.ForeignKey(ContentType, verbose_name=_('content type of the related object'),
                                     null=False, blank=False, on_delete=models.CASCADE)
    object_id = models.TextField(verbose_name=_('ID of the related object'), null=False, blank=False, db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    objects = RelatedObjectManager.from_queryset(RelatedObjectQueryset)()

    class Meta:
        abstract = True
        ordering = ('-created_at',)

    def __str__(self):
        return str(self.content_object)


class BaseAbstractTemplate(SmartModel):

    slug = models.SlugField(verbose_name=_('slug'), max_length=100, null=False, blank=False, editable=False,
                            primary_key=True)
    body = models.TextField(verbose_name=_('message body'), null=True, blank=False)
    is_active = models.BooleanField(null=False, blank=False, default=True, verbose_name=_('is active'))
    is_allowed_duplicate_messages = models.BooleanField(null=False, blank=False, default=True,
                                                        verbose_name=_('Duplicate messages are allowed'))

    def _update_context_data(self, context_data, recipient):
        return context_data

    def render_text_template(self, text, context_data, recipient):
        context_data = self._update_context_data(context_data, recipient)
        return Template(text).render(Context(context_data))

    def render_body(self, context_data, recipient=None):
        return self.render_text_template(self.get_body(), context_data, recipient)

    def clean_body(self, context_data=None):
        try:
            self.render_body(context_data or {})
        except (TemplateSyntaxError, TemplateDoesNotExist) as ex:
            raise ValidationError(_('Error during template body rendering: "{}"').format(ex))

    def get_body(self):
        return self.body

    def can_send_for_object(self, related_objects):
        return (
            (
                self.is_allowed_duplicate_messages
                or not self.exist_duplicate_messages(related_objects)
            )
            and (
                not hasattr(self, 'disallowed_objects')
                or related_objects is None
                or not self.disallowed_objects.filter_from_related_objects(*related_objects).exists()
            )
        )

    def exist_duplicate_messages(self, related_objects):
        if related_objects:
            qs = self.get_controller().model.objects.filter(template=self)
            for obj in related_objects:
                qs = qs.filter(
                    related_objects__object_id=obj.pk,
                    related_objects__content_type=ContentType.objects.get_for_model(obj)
                )
            return qs.exists()
        else:
            return False

    def can_send(self, recipient, related_objects):
        return self.is_active and self.can_send_for_object(related_objects)

    def get_controller(self):
        raise NotImplementedError

    def send(self, recipient, context_data, related_objects=None, tag=None, **kwargs):
        if self.can_send(recipient, related_objects):
            return self.get_controller().send(
                recipient=recipient,
                content=self.render_body(context_data, recipient),
                related_objects=related_objects,
                tag=tag,
                template=self,
                **kwargs,
            )
        else:
            return None

    def __str__(self):
        return self.slug

    class Meta:
        abstract = True
        ordering = ('-created_at',)



import logging
from datetime import timedelta

from chamber.exceptions import PersistenceException
from django.utils.timezone import now

from pymess.backend import BaseBackend, BaseController
from pymess.backend import send as _send
from pymess.backend import send_template as _send_template
from pymess.config import (
    CONTROLLER_TYPES, get_dialer_template_model, get_supported_backend_paths, is_turned_on_dialer_batch_sending,
    settings
)
from pymess.models import DialerMessage


LOGGER = logging.getLogger(__name__)


class DialerController(BaseController):
    """Controller class for dialer delegating message to correct dialer backend"""

    model = DialerMessage
    backend_type_name = CONTROLLER_TYPES.DIALER

    class DialerSendingError(Exception):
        pass

    def get_batch_size(self):
        return settings.DIALER_BATCH_SIZE

    def get_batch_max_seconds_to_send(self):
        return settings.DIALER_BATCH_MAX_SECONDS_TO_SEND

    def get_initial_dialer_state(self, recipient):
        """
        returns initial state for logged dialer instance.
        :param recipient: phone number of the recipient
        """
        return self.model.STATE.WAITING

    def create_message(self, recipient, content=None, related_objects=None, tag=None, template=None, is_autodialer=True,
                       priority=settings.DEFAULT_MESSAGE_PRIORITY, **kwargs):
        """
        Create dialer message which will be logged in the database (content is not needed for this).
        :param recipient: phone number of the recipient
        :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
        :param tag: string mark that will be saved with the message
        :param template: template object from which content of the message was created
        :param is_autodialer: True if it's a autodialer call otherwise False
        :param priority: priority of sending message 1 (highest) to 3 (lowest)
        :param kwargs: extra attributes that will be saved with the message
        """
        extra_data = kwargs.pop('extra_data', {})
        try:
            return super().create_message(
                recipient=recipient,
                content=content,
                related_objects=related_objects,
                tag=tag,
                template=template,
                state=self.get_initial_dialer_state(recipient),
                is_autodialer=is_autodialer,
                priority=priority,
                extra_data=kwargs,
                **self.get_backend(recipient).get_extra_message_kwargs(),
            )
        except PersistenceException as ex:
            raise self.DialerSendingError(str(ex))

    def bulk_check_dialer_status(self):
        """
        Method that finds messages that are not in the final state which were not sent and updates their states.
        """
        messages_to_check = self.model.objects.filter(
            is_final_state=False,
            sent_at__isnull=False,
            backend__in=get_supported_backend_paths(self.backend_type_name),
            created_at__gte=now() - timedelta(minutes=settings.DIALER_IDLE_MESSAGES_TIMEOUT_MINUTES),
        )
        if messages_to_check.exists():
            for backend, messages_for_backend in self._get_backend_messages_map(messages_to_check).items():
                backend._update_dialer_states(messages_for_backend)

    def is_turned_on_batch_sending(self):
        return is_turned_on_dialer_batch_sending()


class DialerBackend(BaseBackend):
    """
    Base class for dialer backend containing implementation of concrete dialer service that
    is used for automatic call to selected phone number.
    """

    def get_batch_max_number_of_send_attempts(self):
        return settings.DIALER_BATCH_MAX_NUMBER_OF_SEND_ATTEMPTS

    def get_retry_sending(self):
        return settings.DIALER_RETRY_SENDING and is_turned_on_dialer_batch_sending()

    def _update_dialer_states(self, messages):
        """
        If dialer sender provides check dialer delivery this method can be overridden.
        :param messages: messages which state will be updated
        """
        raise NotImplementedError('Check dialer state is not supported with the backend')


def send_template(recipient, slug, context_data, related_objects=None, tag=None, send_immediately=False):
    """
    Helper for building and sending dialer message from a template.
    :param recipient: phone number of the recipient
    :param slug: slug of a dialer template
    :param context_data: dict of data that will be sent to the template renderer
    :param related_objects: list of related objects that will be linked with the dialer message using generic
        relation
    :param tag: string mark that will be saved with the message
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :return: dialer message object or None if template cannot be sent
    """
    return _send_template(
        recipient,
        slug,
        context_data,
        related_objects,
        tag,
        template_model=get_dialer_template_model(),
        send_immediately=send_immediately
    )


def send(recipient, content, related_objects=None, tag=None, send_immediately=False, **kwargs):
    """
    Helper for sending dialer message.
    :param recipient: phone number of the recipient
    :param content: text content of the messages
    :param related_objects:
    :param tag: string mark that will be saved with the message
    :param kwargs: extra attributes that will be stored with messages
    :param send_immediately: publishes the message regardless of the `is_turned_on_batch_sending` result
    :return: True if dialer was successfully sent or False if message is in error state
    """
    return _send(
        recipient,
        content,
        related_objects,
        tag,
        message_controller=DialerController(),
        send_immediately=send_immediately,
        **kwargs
    )



from chamber.utils import remove_accent
from chamber.utils.datastructures import ChoicesNumEnum
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from pymess.config import settings
from pymess.utils import normalize_phone_number

from .common import BaseAbstractTemplate, BaseMessage, BaseRelatedObject


__all__ = (
    'OutputSMSMessage',
    'OutputSMSRelatedObject',
    'AbstractSMSTemplate',
    'SMSTemplate',
    'SMSTemplateDisallowedObject',
)


class OutputSMSMessage(BaseMessage):

    STATE = ChoicesNumEnum(
        ('WAITING', _('waiting'), 1),
        ('UNKNOWN', _('unknown'), 2),
        ('SENDING', _('sending'), 3),
        ('SENT', _('sent'), 4),
        ('ERROR_UPDATE', _('error message update'), 5),
        ('DEBUG', _('debug'), 6),
        ('DELIVERED', _('delivered'), 7),
        ('ERROR', _('error'), 8),
        ('ERROR_RETRY', _('error retry'), 9),
    )

    content = models.TextField(verbose_name=_('content'), null=False, blank=False, max_length=700)
    template = models.ForeignKey(settings.SMS_TEMPLATE_MODEL, verbose_name=_('template'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='output_sms_messages')
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False,
                                db_index=True)
    sender = models.CharField(verbose_name=_('sender'), null=True, blank=True, max_length=20)

    class Meta(BaseMessage.Meta):
        verbose_name = _('output SMS')
        verbose_name_plural = _('output SMS')

    def clean_recipient(self):
        self.recipient = normalize_phone_number(force_text(self.recipient))

    def clean_content(self):
        if not settings.SMS_USE_ACCENT:
            self.content = str(remove_accent(str(self.content)))

    @property
    def failed(self):
        return self.state in {self.STATE.ERROR, self.STATE.ERROR_RETRY}


class OutputSMSRelatedObject(BaseRelatedObject):

    output_sms_message = models.ForeignKey(OutputSMSMessage, verbose_name=_('output SMS message'), null=False,
                                           blank=False, on_delete=models.CASCADE, related_name='related_objects')

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('related object of a SMS message')
        verbose_name_plural = _('related objects of SMS messages')


class AbstractSMSTemplate(BaseAbstractTemplate):

    def get_controller(self):
        from pymess.backend.sms import SMSController
        return SMSController()

    class Meta(BaseAbstractTemplate.Meta):
        abstract = True
        verbose_name = _('SMS template')
        verbose_name_plural = _('SMS templates')


class SMSTemplate(AbstractSMSTemplate):
    pass


class SMSTemplateDisallowedObject(BaseRelatedObject):

    template = models.ForeignKey(
        verbose_name=_('template'),
        to=SMSTemplate,
        null=True, blank=True,
        on_delete=models.CASCADE,
        related_name='disallowed_objects',
        db_index=True
    )

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('disallowed object of an SMS template')
        verbose_name_plural = _('disallowed objects of SMS templates')



import json
import logging

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from django.utils.decorators import method_decorator
from django.utils.timezone import now

from is_core.auth.permissions import AllowAny

from pymess.models import EmailMessage


logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name='dispatch')
class MandrillWebhookView(View):

    permission = AllowAny()

    def head(self, *args, **kwargs):
        return HttpResponse()

    def post(self, request, *args, **kwargs):
        try:
            data = json.loads(request.POST.get('mandrill_events'))
        except TypeError:
            return HttpResponse(status=400)

        for event in data:
            try:
                self.process_event(event)
            except Exception as ex:  # pylint: disable=W0703
                logger.exception(ex)

        return HttpResponse()

    def process_event(self, event_dict):
        message_id = event_dict.get('_id', None)
        if message_id:
            message = EmailMessage.objects.filter(external_id=message_id).first()
            if message:
                message.change_and_save(last_webhook_received_at=now())



from chamber.utils.datastructures import ChoicesNumEnum
from django.db import models
from django.utils.translation import ugettext_lazy as _

from pymess.config import settings

from .common import BaseAbstractTemplate, BaseMessage, BaseRelatedObject

__all__ = (
    'AbstractPushNotificationMessage',
    'AbstractPushNotificationTemplate',
    'PushNotificationMessage',
    'PushNotificationMessageRelatedObject',
    'PushNotificationTemplate',
)


class AbstractPushNotificationMessage(BaseMessage):

    STATE = ChoicesNumEnum(
        ('WAITING', _('waiting'), 1),
        ('SENT', _('sent'), 2),
        ('ERROR', _('error'), 3),
        ('DEBUG', _('debug'), 4),
        ('ERROR_RETRY', _('error retry'), 5),
    )

    template = models.ForeignKey(settings.PUSH_NOTIFICATION_TEMPLATE_MODEL, verbose_name=_('template'), blank=True,
                                 null=True, on_delete=models.SET_NULL, related_name='push_notifications')
    state = models.PositiveIntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices,
                                        editable=False, db_index=True)
    heading = models.TextField(verbose_name=_('heading'))
    url = models.URLField(verbose_name=_('URL'), null=True, blank=True)

    class Meta(BaseMessage.Meta):
        abstract = True
        verbose_name = _('push notification')
        verbose_name_plural = _('push notifications')

    @property
    def failed(self):
        return self.state in {self.STATE.ERROR, self.STATE.ERROR_RETRY}


class PushNotificationMessage(AbstractPushNotificationMessage):

    def __str__(self):
        return '{recipient}, {template_slug}, {state}'.format(recipient=self.recipient,
                                                              template_slug=self.template_slug, state=self.state)


class PushNotificationMessageRelatedObject(BaseRelatedObject):

    push_notification_message = models.ForeignKey(PushNotificationMessage, verbose_name=_('push notification message'),
                                                  null=False, blank=False, on_delete=models.CASCADE,
                                                  related_name='related_objects')

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('related object of a push notification message')
        verbose_name_plural = _('related objects of a push notification message')


class AbstractPushNotificationTemplate(BaseAbstractTemplate):

    heading = models.TextField(verbose_name=_('heading'))

    def get_controller(self):
        from pymess.backend.push import PushNotificationController
        return PushNotificationController()

    def send(self, recipient, context_data, related_objects=None, tag=None, **kwargs):
        return super().send(recipient, context_data, related_objects, tag,
                            state=AbstractPushNotificationMessage.STATE.WAITING,
                            heading=self.render_text_template(self.heading, context_data, recipient), **kwargs)

    class Meta(BaseAbstractTemplate.Meta):
        abstract = True
        verbose_name = _('push notification template')
        verbose_name_plural = _('push notification templates')


class PushNotificationTemplate(AbstractPushNotificationTemplate):
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-09-27 15:57
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailmessage',
            name='number_of_send_attempts',
            field=models.PositiveIntegerField(default=0, verbose_name='number of send attempts'),
        ),
    ]



from django.utils import timezone
from pymess.backend.sms import SMSBackend


class DummySMSBackend(SMSBackend):
    """
    Dummy SMS backend used for testing environments. Backend only logs messages to the database.
    """

    def publish_message(self, message):
        self._update_message_after_sending(message, state=message.STATE.DEBUG, sent_at=timezone.now())



import codecs
import json
import os

from django.core import serializers
from django.core.management.base import BaseCommand

from pymess.config import get_email_template_model


class Command(BaseCommand):
    """
    Command dumps e-mails body to the HTML files inside directory. Every HTML file is named with e-mail slug.
    """

    def add_arguments(self, parser):
        parser.add_argument('--directory', dest='directory', type=str, required=True)
        parser.add_argument('--indent', default=0, dest='indent', type=int,
                            help='Specifies the indent level to use when pretty-printing output')

    def dump_template_model(self, directory, indent):
        EmailTemplate = get_email_template_model()
        if EmailTemplate.objects.exists():
            emails_directory = os.path.join(directory)
            if not os.path.exists(emails_directory):
                os.makedirs(emails_directory)
            data = serializers.serialize('python', EmailTemplate.objects.all())
            for obj in data:
                with codecs.open(os.path.join(emails_directory, '{}.html'.format(obj['pk'])), 'w',
                                 encoding='utf-8-sig') as file:
                    file.write(obj['fields']['body'])
                    del obj['fields']['body']

            with codecs.open(os.path.join(emails_directory, 'data.json'), 'w') as file:
                file.write(json.dumps(data, cls=serializers.json.DjangoJSONEncoder, indent=indent))
            self.stdout.write('Stored: "{}" records'.format(len(data)))
        else:
            self.stdout.write('No records')

    def handle(self, *args, **options):
        indent = options.get('indent')
        directory = options.get('directory')
        self.stdout.write('Dumping e-mails')
        self.dump_template_model(directory, indent)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2020-04-05 15:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pymess', '0017_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='dialermessage',
            name='retry_sending',
            field=models.BooleanField(default=False, verbose_name='retry sending'),
        ),
        migrations.AddField(
            model_name='emailmessage',
            name='retry_sending',
            field=models.BooleanField(default=False, verbose_name='retry sending'),
        ),
        migrations.AddField(
            model_name='outputsmsmessage',
            name='retry_sending',
            field=models.BooleanField(default=False, verbose_name='retry sending'),
        ),
        migrations.AddField(
            model_name='pushnotificationmessage',
            name='retry_sending',
            field=models.BooleanField(default=False, verbose_name='retry sending'),
        ),
    ]



import os

from setuptools import setup, find_packages

from pymess.version import get_version


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="django-pymess",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    version=get_version(),
    description="Pymess is a Django framework for sending messages",
    author='Lubos Matl,Oskar Hollman',
    author_email='matllubos@gmail.com',
    url='https://github.com/druids/django-pymess',
    license='LGPL',
    package_dir={'pymess': 'pymess'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    install_requires=[
        'django>=3.1',
        'django-chamber>=0.5.8',
        'attrdict>=2.0.1',
        'beautifulsoup4==4.8.0',
    ],
    zip_safe=False
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-04-01 09:45
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('pymess', '0006_auto_20190322_1712'),
    ]

    operations = [
        migrations.CreateModel(
            name='DialerTemplateDisallowedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='ID of the related object')),
                ('object_id_int', models.PositiveIntegerField(blank=True, db_index=True, null=True, verbose_name='ID of the related object in int format')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='content type of the related object')),
            ],
            options={
                'verbose_name': 'disallowed object of a dialer template',
                'verbose_name_plural': 'disallowed objects of dialer templates',
                'ordering': ('-created_at',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EmailTemplateDisallowedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='ID of the related object')),
                ('object_id_int', models.PositiveIntegerField(blank=True, db_index=True, null=True, verbose_name='ID of the related object in int format')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='content type of the related object')),
            ],
            options={
                'verbose_name': 'disallowed object of an e-mail template',
                'verbose_name_plural': 'disallowed objects of e-mail templates',
                'ordering': ('-created_at',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SMSTemplateDisallowedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='ID of the related object')),
                ('object_id_int', models.PositiveIntegerField(blank=True, db_index=True, null=True, verbose_name='ID of the related object in int format')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType', verbose_name='content type of the related object')),
            ],
            options={
                'verbose_name': 'disallowed object of an SMS template',
                'verbose_name_plural': 'disallowed objects of SMS templates',
                'ordering': ('-created_at',),
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='smstemplatedisallowedobject',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='disallowed_objects', to='pymess.SMSTemplate', verbose_name='template', null=True, blank=True),
        ),
        migrations.AddField(
            model_name='emailtemplatedisallowedobject',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='disallowed_objects', to='pymess.EmailTemplate', verbose_name='template', null=True, blank=True),
        ),
        migrations.AddField(
            model_name='dialertemplatedisallowedobject',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='disallowed_objects', to='pymess.DialerTemplate', verbose_name='template', null=True, blank=True),
        ),
    ]



from chamber.utils.datastructures import ChoicesNumEnum
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from pymess.config import settings
from pymess.utils import normalize_phone_number

from .common import BaseAbstractTemplate, BaseMessage, BaseRelatedObject

__all__ = (
    'AbstractDialerMessage',
    'AbstractDialerTemplate',
    'DialerMessage',
    'DialerMessageRelatedObject',
    'DialerTemplate',
    'DialerTemplateDisallowedObject',
)


class AbstractDialerMessage(BaseMessage):

    STATE = ChoicesNumEnum(
        # numbers are matching predefined state values in Daktela
        ('WAITING', _('waiting'), -1),
        ('NOT_ASSIGNED', _('not assigned'), 0),
        ('READY', _('ready'), 1),
        ('RESCHEDULED_BY_DIALER', _('rescheduled by dialer'), 2),
        ('CALL_IN_PROGRESS', _('call in progress'), 3),
        ('HANGUP', _('hangup'), 4),
        ('DONE', _('done'), 5),
        ('RESCHEDULED', _('rescheduled'), 6),
        ('ANSWERED_COMPLETE', _('listened up complete message'), 7),
        ('ANSWERED_PARTIAL', _('listened up partial message'), 8),
        ('UNREACHABLE', _('unreachable'), 9),
        ('DECLINED', _('declined'), 10),
        ('UNANSWERED', _('unanswered'), 11),
        ('HANGUP_BY_DIALER', _('unanswered - hangup by dialer'), 12),
        ('HANGUP_BY_CUSTOMER', _('answered - hangup by customer'), 13),
        ('ERROR_UPDATE', _('error message update'), 66),
        ('DEBUG', _('debug'), 77),
        ('ERROR', _('error'), 88),
        ('ERROR_RETRY', _('error retry'), 99),
    )

    template = models.ForeignKey(settings.DIALER_TEMPLATE_MODEL, verbose_name=_('template'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='dialer_messages')
    state = models.IntegerField(verbose_name=_('state'), null=False, blank=False, choices=STATE.choices, editable=False,
                                db_index=True)
    is_autodialer = models.BooleanField(verbose_name=_('is autodialer'), null=False, default=True)
    number_of_status_check_attempts = models.PositiveIntegerField(verbose_name=_('number of status check attempts'),
                                                                  null=False, blank=False, default=0)
    content = models.TextField(verbose_name=_('content'), null=True, blank=True)

    class Meta(BaseMessage.Meta):
        abstract = True
        verbose_name = _('dialer message')
        verbose_name_plural = _('dialer messages')

    def clean(self):
        if self.is_autodialer and not self.content:
            raise ValidationError(_('Autodialer message must contain content.'))
        super().clean()

    def clean_recipient(self):
        self.recipient = normalize_phone_number(force_text(self.recipient))

    @property
    def failed(self):
        return self.state in {self.STATE.ERROR, self.STATE.ERROR_RETRY}


class DialerMessage(AbstractDialerMessage):

    is_final_state = models.BooleanField(verbose_name=_('is final state'), null=False, default=False)

    def __str__(self):
        return '{recipient}, {template_slug}, {state}'.format(
            recipient=self.recipient, template_slug=self.template_slug, state=self.get_state_display(),
        )


class DialerMessageRelatedObject(BaseRelatedObject):
    dialer_message = models.ForeignKey(DialerMessage, verbose_name=_('dialer message'), null=False,
                                       blank=False, on_delete=models.CASCADE, related_name='related_objects')

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('related object of a dialer message')
        verbose_name_plural = _('related objects of dialer messages')


class AbstractDialerTemplate(BaseAbstractTemplate):

    def get_controller(self):
        from pymess.backend.dialer import DialerController
        return DialerController()

    def send(self, recipient, context_data, related_objects=None, tag=None, **kwargs):
        return super().send(recipient, context_data, related_objects, tag, **kwargs)

    class Meta(BaseAbstractTemplate.Meta):
        abstract = True
        verbose_name = _('dialer template')
        verbose_name_plural = _('dialer templates')


class DialerTemplate(AbstractDialerTemplate):
    pass


class DialerTemplateDisallowedObject(BaseRelatedObject):

    template = models.ForeignKey(
        verbose_name=_('template'),
        to=DialerTemplate,
        null=True, blank=True,
        on_delete=models.CASCADE,
        related_name='disallowed_objects',
        db_index=True
    )

    class Meta(BaseRelatedObject.Meta):
        verbose_name = _('disallowed object of a dialer template')
        verbose_name_plural = _('disallowed objects of dialer templates')



from attrdict import AttrDict
from bs4 import BeautifulSoup

import requests

from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.template.loader import render_to_string

from chamber.utils.datastructures import ChoicesNumEnum, Enum

from pymess.backend.sms import SMSBackend
from pymess.models import OutputSMSMessage
from pymess.utils.logged_requests import generate_session
from pymess.config import settings


class SMSOperatorBackend(SMSBackend):
    """
    SMS backend that implements ATS operator service https://www.sms-operator.cz/
    Backend supports check SMS delivery
    """

    class SMSOperatorSendingError(Exception):
        pass

    REQUEST_TYPES = Enum(
        'SMS',
        'DELIVERY_REQUEST',
    )

    TEMPLATES = {
        'base': 'pymess/sms/sms_operator/base.xml',
        REQUEST_TYPES.SMS: 'pymess/sms/sms_operator/sms.xml',
        REQUEST_TYPES.DELIVERY_REQUEST: 'pymess/sms/sms_operator/delivery_request.xml',
    }

    SMS_OPERATOR_STATES = ChoicesNumEnum(
        # SMS states
        ('DELIVERED', _('delivered'), 0),
        ('NOT_DELIVERED', _('not delivered'), 1),
        ('PHONE_NUMBER_NOT_EXISTS', _('number not exists'), 2),

        # SMS not moved to GSM operator
        ('TIMEOUTED', _('timeouted'), 3),
        ('INVALID_PHONE_NUMBER', _('wrong number format'), 4),
        ('ANOTHER_ERROR', _('another error'), 5),
        ('EVENT_ERROR', _('event error'), 6),
        ('SMS_TEXT_TOO_LONG', _('SMS text too long'), 7),

        # SMS with more parts
        ('PARTLY_DELIVERED', _('partly delivered'), 10),
        ('UNKNOWN', _('unknown'), 11),
        ('PARLY_DELIVERED_PARTLY_UNKNOWN', _('partly delivered, partly unknown'), 12),
        ('PARTLY_NOT_DELIVERED_PARTLY_UNKNOWN', _('partly not delivered, partly unknown'), 13),
        ('PARTLY_DELIVERED_PARTLY_NOT_DELIVERED_PARTLY_UNKNOWN',
         _('partly delivered, partly not delivered, partly unknown'), 14),
        ('NOT_FOUND', _('not found'), 15),
    )

    SMS_OPERATOR_STATES_MAPPING = {
        SMS_OPERATOR_STATES.DELIVERED: OutputSMSMessage.STATE.DELIVERED,
        SMS_OPERATOR_STATES.NOT_DELIVERED: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.PHONE_NUMBER_NOT_EXISTS: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.TIMEOUTED: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.INVALID_PHONE_NUMBER: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.ANOTHER_ERROR: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.EVENT_ERROR: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.SMS_TEXT_TOO_LONG: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.PARTLY_DELIVERED: OutputSMSMessage.STATE.ERROR_UPDATE,
        SMS_OPERATOR_STATES.UNKNOWN: OutputSMSMessage.STATE.SENDING,
        SMS_OPERATOR_STATES.PARLY_DELIVERED_PARTLY_UNKNOWN: OutputSMSMessage.STATE.SENDING,
        SMS_OPERATOR_STATES.PARTLY_NOT_DELIVERED_PARTLY_UNKNOWN: OutputSMSMessage.STATE.SENDING,
        SMS_OPERATOR_STATES.PARTLY_DELIVERED_PARTLY_NOT_DELIVERED_PARTLY_UNKNOWN:
            OutputSMSMessage.STATE.SENDING,
        SMS_OPERATOR_STATES.NOT_FOUND: OutputSMSMessage.STATE.ERROR_UPDATE,
    }

    config = AttrDict({
        'URL': 'https://www.sms-operator.cz/webservices/webservice.aspx',
        'UNIQ_PREFIX': '',
        'TIMEOUT': 5,  # 5s
    })

    def _get_extra_sender_data(self):
        return {
            'prefix': self.config.UNIQ_PREFIX,
        }

    def _serialize_messages(self, messages, request_type):
        """
        Serialize SMS messages to the XML
        :param messages: list of SMS messages
        :param request_type: type of the request to the SMS operator
        :return: serialized XML message that will be sent to the SMS operator service
        """
        return render_to_string(
            self.TEMPLATES['base'], {
                'username': self.config.USERNAME,
                'password': self.config.PASSWORD,
                'prefix': str(self.config.UNIQ_PREFIX) + '-',
                'template_type': self.TEMPLATES[request_type],
                'messages': messages,
                'type': 'SMS' if request_type == self.REQUEST_TYPES.SMS else 'SMS-Status',
            }
        )

    def _send_requests(self, messages, request_type, is_sending=False, **change_sms_kwargs):
        """
        Performs the actual POST request for input messages and request type.
        :param messages: list of SMS messages
        :param request_type: type of the request
        :param is_sending: True if method is called after sending message
        :param change_sms_kwargs: extra kwargs that will be stored to the message object
        """
        requests_xml = self._serialize_messages(messages, request_type)
        try:
            resp = generate_session(slug='pymess - SMS operator', related_objects=list(messages)).post(
                self.config.URL,
                data=requests_xml,
                headers={'Content-Type': 'text/xml'},
                timeout=self.config.TIMEOUT
            )
            if resp.status_code != 200:
                raise self.SMSOperatorSendingError(
                    'SMS operator returned invalid response status code: {}'.format(resp.status_code)
                )
            self._update_sms_states_from_response(
                messages, self._parse_response_codes(resp.text), is_sending, **change_sms_kwargs
            )
        except requests.exceptions.RequestException as ex:
            raise self.SMSOperatorSendingError(
                'SMS operator returned returned exception: {}'.format(str(ex))
            )

    def _update_sms_states_from_response(self, messages, parsed_response, is_sending=False, **change_sms_kwargs):
        """
        Higher-level function performing serialization of SMS operator requests, parsing ATS server response and
        updating  SMS messages state according the received response.
        :param messages: list of SMS messages
        :param parsed_response: parsed HTTP response from the SMS operator service
        :param is_sending: True if update is called after sending message
        :param change_sms_kwargs: extra kwargs that will be stored to the message object
        """

        messages_dict = {message.pk: message for message in messages}

        missing_uniq = set(messages_dict.keys()) - set(parsed_response.keys())
        if missing_uniq:
            raise self.SMSOperatorSendingError(
                'SMS operator not returned SMS info with uniq: {}'.format(', '.join(map(str, missing_uniq)))
            )

        extra_uniq = set(parsed_response.keys()) - set(messages_dict.keys())
        if extra_uniq:
            raise self.SMSOperatorSendingError(
                'SMS operator returned SMS info about unknown uniq: {}'.format(', '.join(map(str, extra_uniq)))
            )

        for uniq, sms_operator_state in parsed_response.items():
            sms = messages_dict[uniq]
            state = self.SMS_OPERATOR_STATES_MAPPING.get(sms_operator_state)
            error = (
                self.SMS_OPERATOR_STATES.get_label(sms_operator_state)
                if state == OutputSMSMessage.STATE.ERROR_UPDATE else None
            )
            if is_sending:
                if error:
                    self._update_message_after_sending_error(
                        sms,
                        state=state,
                        error=error,
                        extra_sender_data={'sender_state': sms_operator_state},
                        **change_sms_kwargs
                    )
                else:
                    self._update_message_after_sending(
                        sms,
                        state=state,
                        extra_sender_data={'sender_state': sms_operator_state},
                        **change_sms_kwargs
                    )
            else:
                self._update_message(
                    sms,
                    state=state,
                    error=error,
                    extra_sender_data={'sender_state': sms_operator_state},
                    **change_sms_kwargs
                )

    def publish_message(self, message):
        try:
            self._send_requests(
                [message],
                request_type=self.REQUEST_TYPES.SMS,
                is_sending=True,
                sent_at=timezone.now()
            )
        except self.SMSOperatorSendingError as ex:
            self._update_message_after_sending_error(
                message,
                state=OutputSMSMessage.STATE.ERROR,
                error=str(ex)
            )
        except requests.exceptions.RequestException as ex:
            self._update_message_after_sending_error(
                message,
                error=str(ex)
            )
            # Do not re-raise caught exception. Re-raise exception causes transaction rollback (lost of information
            # about exception).

    def publish_messages(self, messages):
        self._send_requests(messages, request_type=self.REQUEST_TYPES.SMS, is_sending=True, sent_at=timezone.now())

    def _parse_response_codes(self, xml):
        """
        Finds all <dataitem> tags in the given XML and returns a mapping "uniq" -> "response code" for all SMS.
        In case of an error, the error is logged.
        :param xml: XML from the SMS operator response
        :return: dictionary with pair {SMS uniq: response status code}
        """

        soup = BeautifulSoup(xml, 'html.parser')

        return {int(item.smsid.string.lstrip(self.config.UNIQ_PREFIX + '-')): int(item.status.string)
                for item in soup.find_all('dataitem')}

    def update_sms_states(self, messages):
        self._send_requests(messages, request_type=self.REQUEST_TYPES.DELIVERY_REQUEST)



import logging

from chamber.utils.transaction import atomic_with_signals
from django.core.management.base import BaseCommand, CommandError
from django.db import DatabaseError

from pymess.backend.dialer import DialerController
from pymess.backend.emails import EmailController
from pymess.backend.push import PushNotificationController
from pymess.backend.sms import SMSController

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Command for sending messages in batch.
    """

    controllers = {
        'email': EmailController(),
        'push-notification': PushNotificationController(),
        'dialer': DialerController(),
        'sms': SMSController()
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.touched_message_pks = set()
        self.send_message_pks = set()
        self.failed_message_pks = set()

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--type', action='store', dest='type', default='email',
                            help='Tells Django what type of messages should be send '
                                 '(email/push-notification/dialer/sms).')

    @atomic_with_signals
    def _send_message(self, controller):
        message = controller.get_waiting_or_retry_messages().exclude(
            pk__in=self.touched_message_pks
        ).select_for_update(
            nowait=True
        ).order_by('priority', 'created_at').first()
        if not message or not controller.is_turned_on_batch_sending():
            return False

        self.touched_message_pks.add(message.pk)
        try:
            if controller.publish_or_retry_message(message):
                self.send_message_pks.add(message.pk)
            else:
                self.failed_message_pks.add(message.pk)
        except Exception as ex:
            # Rollback should not be applied for already send e-mails
            logger.exception(ex)
            self.failed_message_pks.add(message.pk)
        return True

    def _print_result(self, title, message_pks):
        if message_pks:
            self.stdout.write('{}: {} ({})'.format(title, len(message_pks), ', '.join((str(pk) for pk in message_pks))))
        else:
            self.stdout.write('{}: {}'.format(title, len(message_pks)))

    def handle(self, type, *args, **options):
        controller = self.controllers[type]
        if not controller.is_turned_on_batch_sending():
            raise CommandError('Batch sending is turned off')

        try:
            for _ in range(controller.get_batch_size()):
                if not self._send_message(controller):
                    break
            self._print_result('sent messages', self.send_message_pks)
            self._print_result('failed messages', self.failed_message_pks)
        except DatabaseError:
            self.stdout.write('messages are already locked')
